{-# LANGUAGE TypeFamilies,
             AllowAmbiguousTypes,
             GeneralizedNewtypeDeriving #-}

module CmRDT where

import Numeric.Natural
import VectorClock

class CmRDT a where
  type Type a
  type Operation a
  type Effect a
  query :: a -> Type a
  prepare :: Operation a -> a -> Effect a
  update :: Effect a -> a -> a

newtype Number = Number Integer deriving (Ord,Num,Eq,Show,Real,Integral,Enum)

instance CmRDT Number where
  type Type Number = Integer
  type Operation Number = Integer
  type Effect Number = Integer
  query (Number n) = n
  prepare = const
  update = (+) . fromIntegral


data LWWRegister a = LWWRegister {
  siteId :: Int,
  clock ::  VectorClock,
  value :: a } deriving (Show,Eq)

instance CmRDT (LWWRegister a) where
  type Type (LWWRegister a) = a
  type Operation (LWWRegister a) = a
  type Effect (LWWRegister a) = (VectorClock, a)
  query (LWWRegister id t a) = a
  prepare a' (LWWRegister id t a) = (inc id t, a)  
  update (t',a') (LWWRegister id t a) = 
    if t' > t then (LWWRegister id t'' a') else (LWWRegister id t'' a)
      where t'' = merge t t'