{-# LANGUAGE TypeFamilies,
             FlexibleInstances,
             AllowAmbiguousTypes,
             GeneralizedNewtypeDeriving #-}

module CvRDT where

import Data.List (nub,union,(\\),sort)
import Numeric.Natural

class CvRDT a where    
  type Type a
  type Operation a
  query :: a -> Type a
  update :: Operation a -> a -> a
  merge :: a -> a -> a

newtype Counter = Counter Natural deriving (Ord,Num,Eq,Show,Real,Integral,Enum)

instance CvRDT Counter where  
  type Type Counter = Natural
  type Operation Counter = ()
  query (Counter v) = v
  update = const (+1)
  merge a b = max a b

data PNCounter = PNCounter Counter Counter deriving (Eq,Show)
instance Ord PNCounter where
  (PNCounter p n) <= (PNCounter p' n') = p <= p' && n <= n'

instance CvRDT PNCounter where  
  type Type PNCounter = Integer
  type Operation PNCounter = Bool
  query (PNCounter p n) = (fromIntegral p) - (fromIntegral n)
  update True (PNCounter p n) = PNCounter (p+1) n
  update False (PNCounter p n) = PNCounter p (n+1)
  merge (PNCounter p1 n1) (PNCounter p2 n2) = PNCounter (max p1 p2) (max n1 n2)

newtype GrowOnlySet a = GrowOnlySet [a] deriving (Show,Eq)
instance Eq a => Ord (GrowOnlySet a) where
  (GrowOnlySet a) <= (GrowOnlySet b) = length a <= length b

instance (Eq a, Ord a) => CvRDT (GrowOnlySet a) where
  type Type (GrowOnlySet a) = [a]
  type Operation (GrowOnlySet a) = a
  query (GrowOnlySet as) = sort as
  update a (GrowOnlySet as) = GrowOnlySet (union as [a])
  merge (GrowOnlySet a) (GrowOnlySet b) = GrowOnlySet $ union a b

data TwoPhaseSet a = TwoPhaseSet (GrowOnlySet a) (GrowOnlySet a) deriving (Eq,Show)
instance (Eq a) => Ord (TwoPhaseSet a) where 
  (TwoPhaseSet p n) <= (TwoPhaseSet p' n') = p <= p' && n <= n'

instance (Ord a, Eq a) => CvRDT (TwoPhaseSet a) where
  type Type (TwoPhaseSet a) = [a]
  type Operation (TwoPhaseSet a) = (a,Bool)
  query (TwoPhaseSet p n) = sort (query p \\ query n)
  update (e,True) (TwoPhaseSet p n) = TwoPhaseSet (update e p) n
  update (e,False) (TwoPhaseSet p n) = TwoPhaseSet p (update e n)
  merge (TwoPhaseSet a1 d1) (TwoPhaseSet a2 d2) = TwoPhaseSet (merge a1 a2) (merge d1 d2)