{-# LANGUAGE AllowAmbiguousTypes,
             ScopedTypeVariables,
             FlexibleContexts,
             FlexibleInstances #-}

module Tests.CmRDT where

import CmRDT
import Test.Tasty
import qualified Test.Tasty.QuickCheck as QC
import Test.Tasty.QuickCheck ((==>))
import Data.List (nub)
import Type.Reflection
import VectorClock

test :: forall a . (Typeable a, 
                    Eq (Type a), 
                    CmRDT a, 
                    QC.Arbitrary a, 
                    QC.Arbitrary (Replicas a), 
                    QC.Arbitrary (Operation a), 
                    Show a,
                    Show (Operation a)) => TypeRep a -> TestTree
test rep = testGroup (show rep) [
    QC.testProperty "Commutativity" $ commutativity
    --, QC.testProperty "Causality" $ causality
  ] where commutativity :: Operation a -> Operation a -> a -> Bool
          commutativity a b s = 
            query (update (prepare b as) as) == query (update (prepare a bs) bs)  
            where as = update (prepare a s) s
                  bs = update (prepare b s) s
{-}
          causality :: Replicas a -> [(Int,Operation a)] -> Property
          causality (Replicas reps) operations = forAll (listOf infiniteList) $ \orders -> 
           ((==1) . length) map (run reps msgs) orders 
             where msgs = map (\(rep,i) -> map snd $ filter (==i . fst) operations) (zip reps [0..])
                   run states msgs order = if (all null state)  -}
          {-causality :: Replicas a -> Operation a -> Operation a -> Bool
          causality (Replicas s1 s2) a b = ((==1) . length) $ nub $ map query results -- check all the following are equivalent                
              where results = concatMap (\(eff,ss) -> map (update eff) ss) pairs-}

   
newtype Replicas a = Replicas [a] deriving Show

instance QC.Arbitrary (Replicas Number) where
  arbitrary = QC.sized (\i -> fmap (Replicas . replicate i) QC.arbitrary)

instance QC.Arbitrary a => QC.Arbitrary (Replicas (LWWRegister a)) where
  arbitrary = QC.sized $ \i -> do
    value <- QC.arbitrary
    let t = VectorClock (replicate 0 i)        
    return $ Replicas (map (\id -> LWWRegister id t value) [0..i])

instance QC.Arbitrary a => QC.Arbitrary (LWWRegister a) where
  arbitrary = do
    a <- QC.arbitrary
    size <- QC.choose (2,8)
    x <- QC.vectorOf size QC.arbitrary
    id <- QC.choose (0, length x - 1)
    return $ LWWRegister id (VectorClock x) a



instance QC.Arbitrary Number where 
  arbitrary = fmap Number QC.arbitrary
  
tests = testGroup "CmRDTs" [
    test (typeRep :: TypeRep Number),
    test (typeRep :: TypeRep (LWWRegister Int))
  ]