{-# LANGUAGE AllowAmbiguousTypes,
             ScopedTypeVariables,
             FlexibleContexts #-}

module Tests.CvRDT where

import CvRDT
import Test.Tasty
import qualified Test.Tasty.QuickCheck as QC
import Data.List (nub)
import Type.Reflection

test :: forall a . (Typeable a, QC.Arbitrary a, Show a, CvRDT a, Eq (Type a), Ord a, QC.Arbitrary (Operation a), Show (Operation a)) => TypeRep a -> TestTree
test rep = testGroup (show rep) [
    QC.testProperty "Commutativity" $ commutativity,
    QC.testProperty "Idempotence" $ idempotence,
    QC.testProperty "Associativity" $ associativity,
    QC.testProperty "Monotonicity" $ monotonicity
  ] where commutativity :: a -> a -> Bool
          commutativity a b = query (merge a b) == query (merge b a)
          idempotence :: a -> Bool
          idempotence a = query a == query (merge a a)
          associativity :: a -> a -> a -> Bool
          associativity a b c = query (merge a (merge b c)) == query (merge (merge a b) c)
          monotonicity :: a -> Operation a -> Bool
          monotonicity a op = update op a >= a
   
instance QC.Arbitrary Counter where 
  arbitrary = fmap (Counter . fromInteger . abs) QC.arbitrary

instance QC.Arbitrary PNCounter where 
  arbitrary = do
    p <- QC.arbitrary
    n <- QC.arbitrary
    return $ PNCounter p n
      
instance (Eq a, QC.Arbitrary a) => QC.Arbitrary (GrowOnlySet a) where
  arbitrary = fmap (GrowOnlySet . nub) QC.arbitrary

instance (Eq a, QC.Arbitrary a) => QC.Arbitrary (TwoPhaseSet a) where
  arbitrary = do
    p <- QC.arbitrary
    n <- QC.arbitrary
    return $ TwoPhaseSet p n
  

tests = testGroup "CvRDTs" [
    test (typeRep :: TypeRep Counter),
    test (typeRep :: TypeRep PNCounter),
    test (typeRep :: TypeRep (GrowOnlySet Int)),
    test (typeRep :: TypeRep (TwoPhaseSet Int))
  ]